#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ###########################################################
# Auslagerung einiger Funktionen, die nix mit wxPython
# zu tun haben.
#
# Detlev Ahlgrimm, 2017

import time
import math
import os
from lxml import etree  # python-lxml-3.3.5-2.1.4.x86_64
from datetime import datetime, tzinfo, timedelta
from calendar import timegm
import requests

# ###########################################################
# Quelle: https://aboutsimon.com/blog/2013/06/06/Datetime-hell-Time-zone-aware-to-UNIX-timestamp.html
class UTC(tzinfo):
    """UTC"""
    def utcoffset(self, dt):
        return timedelta(0)
    def tzname(self, dt):
        return "UTC"
    def dst(self, dt):
        return timedelta(0)


class OSMhelper():
  def __init__(self):
    pass

  # ###########################################################
  # Liefert den UNIX-Timestamp für einen Timestamp-String in
  # UTC.
  # Erweitert mit: http://stackoverflow.com/a/33148723/3588613
  def getTimestampStringAsUTC(self, strg):
    if "." in strg:
      datetime_obj = datetime.strptime(strg, '%Y-%m-%dT%H:%M:%S.%fZ')
    else:
      datetime_obj = datetime.strptime(strg, '%Y-%m-%dT%H:%M:%SZ')
    datetime_obj = datetime_obj.replace(tzinfo=UTC())
    timestamp = timegm(datetime_obj.timetuple())
    return(timestamp)


  # ######################################################################
  # Liefert zu einem UNIX-Timestamp (UTC, Integer) einen String mit der
  # Ortszeit.
  def getUTCTimestampAsLocal(self, ts):
    return(datetime.fromtimestamp(ts).strftime("%Y.%m.%d %H:%M:%S"))


  # ######################################################################
  # Quelle: https://pypi.python.org/pypi/haversine
  # AVG_EARTH_RADIUS = 6371  - in km
  def haversine(self, point1, point2, miles=False):
    # unpack latitude/longitude
    lat1, lng1 = point1
    lat2, lng2 = point2

    # convert all latitudes/longitudes from decimal degrees to radians
    lat1, lng1, lat2, lng2 = map(math.radians, (lat1, lng1, lat2, lng2))

    # calculate haversine
    lat = lat2 - lat1
    lng = lng2 - lng1
    d = math.sin(lat * 0.5) ** 2 + math.cos(lat1) * math.cos(lat2) * math.sin(lng * 0.5) ** 2
    h = 2 * 6371 * math.asin(math.sqrt(d))
    if miles:
        return h * 0.621371  # in miles
    else:
        return h  # in kilometers


  # ###########################################################
  # Liefert die Rückgabe von "nominatim.openstreetmap.org" zu
  # einer Geo-Koordinate und Zoom-Level.
  def getNominatim(self, (lat, lon), zoom):
    url="http://nominatim.openstreetmap.org/reverse?format=xml&lat=%f&lon=%f&zoom=%d&addressdetails=1"%(lat, lon, zoom)
    text=""
    response=requests.get(url, timeout=1)
    try:
      root=etree.fromstring(response.content)
    except:
      print "response.content= %s\n"%(response.content,),
      return(text)
    for elem in root:
      if elem.tag=="result":
        text=elem.text.replace(", ", ", \n ")
        break
    return(text)


  # ###########################################################
  # Get all points of a straight line.
  # Quelle: http://stackoverflow.com/a/25913345/3588613
  def getLine(self, (x1, y1), (x2, y2)):
    points = []
    issteep = abs(y2-y1) > abs(x2-x1)
    if issteep:
        x1, y1 = y1, x1
        x2, y2 = y2, x2
    rev = False
    if x1 > x2:
        x1, x2 = x2, x1
        y1, y2 = y2, y1
        rev = True
    deltax = x2 - x1
    deltay = abs(y2-y1)
    error = int(deltax / 2)
    y = y1
    ystep = None
    if y1 < y2:
        ystep = 1
    else:
        ystep = -1
    for x in range(x1, x2 + 1):
        if issteep:
            points.append((y, x))
        else:
            points.append((x, y))
        error -= deltay
        if error < 0:
            y += ystep
            error += deltax
    # Reverse the list if the coordinates were reversed
    if rev:
        points.reverse()
    return points


  # ######################################################################
  # GPX-Datei laden, parsen und als Liste von Tracks, bestehend aus Listen
  # von Tupeln ((lat, lon), timestamp), zurückgeben.
  # Bei ignore_trkseg==False werden die Tags <trk> und <trkseg>
  # gleichwertig behandelt - in beiden Fällen wird eine neue Sub-Liste
  # erstellt.
  # Bei ignore_trkseg==True wird das Tag <trkseg> ignoriert und somit der
  # aktuelle fortgesetzt.
  def loadGPX(self, filename, ignore_trkseg=True):
    try:
      with open(filename, "r") as fl:
        filedata=fl.read()
    except:
      return([], [])
    try:
      root=etree.fromstring(filedata)
    except:
      return([], [])
    ns=""
    if None in root.nsmap:
      ns="{"+root.nsmap[None]+"}"
    tracks=list()
    wpts=list()
    if etree.iselement(root) and root.tag==ns+"gpx":
      for trk in root:
        if etree.iselement(trk) and trk.tag==ns+"wpt":
          wptLL=(float(trk.get("lat")), float(trk.get("lon")))
          wptData=dict()
          for e in trk:
            if etree.iselement(e) and e.tag==ns+"name":
              wptData.update({"name":e.text})
            if etree.iselement(e) and e.tag==ns+"desc":
              wptData.update({"desc":e.text})
            if etree.iselement(e) and e.tag==ns+"type":
              wptData.update({"type":e.text})
          wpts.append((wptLL, wptData))
        if ignore_trkseg: track=list()
        if etree.iselement(trk) and trk.tag==ns+"trk":
          for trkseg in trk:
            if etree.iselement(trkseg) and trkseg.tag==ns+"trkseg":
              if not ignore_trkseg: track=list()
              for trkpt in trkseg:
                if etree.iselement(trkpt) and trkpt.tag==ns+"trkpt":
                  t=0
                  for e in trkpt:
                    if etree.iselement(e) and e.tag==ns+"time":
                      t=self.getTimestampStringAsUTC(e.text)
                      break
                  track.append(((float(trkpt.get("lat")), float(trkpt.get("lon"))), t))
              if not ignore_trkseg and len(track)>0:
                tracks.append(track)
        if ignore_trkseg and len(track)>0:
          tracks.append(track)
    if len(tracks)>0:
      self.file_loaded=True
    return(tracks, wpts)


  # ######################################################################
  # Liefert für einen Track (als Liste aus Tupeln ((lat, lon), timestamp))
  # die Summe der Abschitts-Entfernungen in Kilometern. 
  def getTrackLength(self, track):
    dist=0
    ptl, tm=track[0]
    for pt, tm in track:
      dist+=self.haversine(ptl, pt)
      ptl=pt
    return(dist)


  # ######################################################################
  # Liefert für einen Track (als Liste aus Tupeln ((lat, lon), timestamp))
  # eine Liste von Punkten mit Start- und Endzeit der Koodinaten, wo
  # mindestens "min_dur_secs" lang kaum Bewegung stattfand. Wobei mit
  # "kaum Bewegung" gemeint ist, dass sämtliche Bewegung mit weniger als
  # "max_ign_spd_kmh" Stundenkilometern erfolgte.
  # Die Rückgabe-Liste enthält Tupel im Format:
  #       (lat, lon, start-timestamp, ende-timestamp)
  # z.B.: (54.7127033, 9.76528, 1469954860, 1469955259)
  def getPausePoints(self, track, min_dur_secs=60, max_ign_spd_kmh=5):
    if len(track)==0:
      return([])
    pause_point=list()      # Rückgabeliste initialisieren
    ptis=-1   # nimmt den Index von "track" eines [potentiellen] Pausen-Start-Punktes auf
    (lato, lono), tmo=track[0]  # Daten des Vorgängersatzes merken bzw. initialisieren
    for idx in range(1, len(track)):
      (lat, lon), tm=track[idx]
      dist=self.haversine((lato, lono), (lat, lon)) # Entfernung in Kilometern
      dur=tm-tmo                                    # Dauer in Sekunden
      if dur==0:  # da haben wohl zwei Sätze die selbe Zeit....
        continue  # ...sowas will man nicht sehen... -> ignorieren
      spd=dist/(dur/3600.0)                         # Geschwindigkeit in km/h
      lato, lono, tmo=lat, lon, tm
      if ptis>=0: # wenn gerade eine potentielle Pause läuft
        if spd>max_ign_spd_kmh: # ...und diese jetzt zuende wäre
          # Track-Punkt: ((lat, lon), timestamp)  -> lat=[0][0], lon=[0][1], timestamp=[1]
          if (track[idx-1][1]-track[ptis][1])>=min_dur_secs: # ...und die Pause lang genug war
            mid_idx=ptis+((idx-1)-ptis)/2 # Index in "track" der Pausen-Mitte bestimmen
            pause_point.append((track[mid_idx][0][0], track[mid_idx][0][1], track[ptis][1], track[idx-1][1]))
          ptis=-1 # die Pause ist jetzt in jedem Fall zuende
      elif ptis<0 and spd<max_ign_spd_kmh:  # wenn neuer potentieller Pause-Punkt gefunden wird...
        ptis=idx  # ...dann dessen Index in "track" vermerken
    return(pause_point)
    


# ######################################################################
# Tests
if __name__=="__main__":
  c=OSMhelper()
  trks=c.loadGPX("/home/dede/wd/xml_tracker/2016/2016_09_25_12_39_25.gpx")

  for trk in trks:
    print c.getPausePoints(trk)

  if False:
    print "-"*60
    pause_point=list()
    lato=lono=tmo=None
    pp_first=pp_last=None
    avg_spd=0.0
    spd_cnt=0
    max_spd=0.0
    for (lat, lon), tm in trk:
      if lato is None:
        dist=0
        dur=0
        lato, lono=lat, lon
        tmo=tm
        spd=0
      else:
        dist=c.haversine((lato, lono), (lat, lon))
        dur=tm-tmo
        spd=dist/(dur/3600.0)
        lato, lono, tmo=lat, lon, tm
      max_spd=max(max_spd, spd)
      if spd>2:
        avg_spd+=spd
        spd_cnt+=1
      if pp_first is None and spd<5:
        pp_first=(lat, lon, tm)
      if pp_first is not None:
        if spd>5:
          if (tm-pp_first[2])>120:
            pause_point.append((pp_first, pp_last))
          pp_first=None
        else:
          pp_last=(lat, lon, tm)
      print "%14.11f %14.11f %11d % 4dm % 3ds %5.1dkm/h"%(lat, lon, tm, dist*1000, dur, spd)
    
    print "max. speed=%4.1fkm/h"%(max_spd)
    print "avg. speed=%4.1fkm/h"%(avg_spd/spd_cnt)
    print pause_point
