#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ###########################################################
# Eine Klasse zum Laden und cachen von OpenStreetMap-Tiles.
#
# Detlev Ahlgrimm, 2017
#
# Eine Erweiterungsmöglichkeit wäre die Abfrage des Erstellungsdatums von
# jedem Tile im Cache und Neuanforderung bei einem Alter >1 Jahr.
#
#
# 08.04.2017  erste Version
# 10.04.2017  Umstellung auf asynchrones Laden der Tiles
# 14.04.2017  self.status eingebaut, damit eine Unterscheidung zwischen
#             "laden unmöglich" und "laden fehlgeschlagen" erfolgen kann.
# 23.04.2017  "force_reload" und __removeTile() zugefügt
# 24.04.2017  RAMcache zugefügt, Löschbedingung für self.already_requested erweitert
#


import wx               # python-wxWidgets-2.8.12.1-10.4.1.x86_64
from PIL import Image   # python-Pillow-2.8.1-3.6.1.x86_64 
# leider taugen die Image-Funktionen von wxPython nix bzw. melden Bild-Fehler per Popup
import requests         # python-requests-2.4.1-3.1.noarch
from io import BytesIO
import os
import threading
from Queue import Queue

DEBUG=True
CACHE_MAX_TILES=1000

class TileCache():
  def __init__(self, cache_base_dir, tile_url):
    self.cache_base_dir=cache_base_dir
    if self.cache_base_dir is None:
      self.cacheRAM=RAMcache()
    self.tile_url=tile_url
    self.queue=Queue()      # Auftrags-Übergabe für die Threads
    self.running_count=0    # wird pro arbeitendem Thread hochgezählt und wieder runter, wenn er fertig ist
    self.running_lock=threading.Lock()  # ein Lock() für Schreibzugriffe auf "self.running_count"
    self.num_worker_threads=35    # Anzahl Worker-Threads
    self.request_timeout=2        # Maximal Wartezeit in Sekunden, bevor Ladevorgang abgebrochen wird
    self.already_requested=list() # Liste aller aktuell zu holenden Tiles
    self.status={"GOOD":0, "WAIT":1, "FAILED":2, "INVALID":3}
    for i in range(self.num_worker_threads):
      worker=threading.Thread(target=self.__getTileAsyncFromWeb, name="getTile("+str(i)+")")
      worker.setDaemon(True)
      worker.start()


  # ######################################################################
  # Liefert das OSM-Tile(x, y, z) mit Status als Bitmap oder None.
  def getTile(self, x, y, z, force_reload=False):
    status, image=self.getTileImg(x, y, z, force_reload)
    if image is not None:
      w, h=image.size                                     # Tile-Größe bestimmen
      bitmap=wx.BitmapFromBuffer(w, h, image.tostring())  # nach Bitmap wandeln
      return((status, bitmap))                            # und zurückliefern
    return((status, None))

    
  # ######################################################################
  # Liefert das OSM-Tile(x, y, z) mit Status als Image oder None.
  def getTileImg(self, x, y, z, force_reload=False):
    if self.queue.empty() and self.running_count==0:  # wenn queue leer ist und alle Threads idlen...
      self.already_requested=list()                   # ...können die alten Merker weg
    if self.cache_base_dir is not None:           # wenn im Dateisystem gecached werden soll
      fldr=os.path.join(self.cache_base_dir, str(z), str(x))  # Pfad bauen
      fl=os.path.join(fldr, "%d.png"%(y,))                    # Tile-Name samt Pfad bauen
      if force_reload:          # wenn "reload" angefordert wurde...
        self.__removeTile(fl)   # Tile vorsichtshalber löschen, damit es auch trotz Timeout sicher "reloaded" wird
      if os.path.exists(fl) and not force_reload:   # wenn Tile schon im Cache existiert
        try:
          img=Image.open(fl)                        # Tile aus Cache liefern
        except:
          return((self.status["FAILED"], None))     # bei Fehler später nochmal probieren
      else:
        if (x, y, z) not in self.already_requested: # wenn Anfrage für Tile noch nicht offen ist
          if x<0 or x>=2**z or y<0 or y>=2**z:      # wenn Tile nicht existiert bzw. illegal ist
            # http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Zoom_levels
            return((self.status["INVALID"], None))  # als ungültiges Tile markieren
          self.queue.put((x, y, z))                 # asynchrones Laden des Tiles beauftragen
          self.already_requested.append((x, y, z))  # und Auftrag merken, um ihn nur einmal abzusetzen
        return((self.status["WAIT"], None))         # Nachlieferungs-Meldung
    else:                                         # wenn im RAM gecached werden soll
      img=self.cacheRAM.getFromRAM(x, y, z)       # alles analog zu oben
      if img is not None and not force_reload:
        pass
      else:
        if (x, y, z) not in self.already_requested:
          if x<0 or x>=2**z or y<0 or y>=2**z:
            return((self.status["INVALID"], None))
          self.queue.put((x, y, z))
          self.already_requested.append((x, y, z))
        return((self.status["WAIT"], None))
    try:
      image=img.convert('RGB')                    # den convert() braucht der wx.BitmapFromBuffer()
    except:
      return((self.status["FAILED"], None))       # wenn Fehlschlag - später nochmal probieren
    return((self.status["GOOD"], image))          # und zurückliefern
#    w, h=image.size                               # Tile-Größe bestimmen
#    bitmap=wx.BitmapFromBuffer(w, h, image.tostring())  # nach Bitmap wandeln
#    return((self.status["GOOD"], bitmap))         # und zurückliefern


  # ######################################################################
  # Löscht das Tile "filename" aus dem Dateisystem.
  def __removeTile(self, filename):
    try:
      os.remove(filename)
    except:
      pass


  # ######################################################################
  # Löscht alle anstehenden Requests, die nicht schon von einem
  # Worker-Thread angenommen wurden.
  def flushQueue(self):
    while not self.queue.empty():
      self.queue.get()
    self.already_requested=list()


  # ######################################################################
  # Liefert die Anzahl der noch unbeantworteten Tile-Requests.
  def getQueueLength(self):
    return(self.queue.qsize()+self.running_count)


  # ######################################################################
  # Liefert die Textstrings und Fehlernummern, die bei Fehler statt
  # Bitmaps geliefert werden können als Dictionary.
  def getStatusDict(self):
    return(self.status)


  # ######################################################################
  # Worker-Thread, um ein OSM-Tile zu laden und lokal zu speichern.
  def __getTileAsyncFromWeb(self):
    while True:
      job=self.queue.get()
      self.__incRunningCount()
      x, y, z=job
      s, img=self.__getTileFromWeb(x, y, z)
      if s==self.status["GOOD"]:
        if self.cache_base_dir is None:
          self.cacheRAM.putIntoRAM(x, y, z, img)
        else:
          fldr=os.path.join(self.cache_base_dir, str(z), str(x))
          fl=os.path.join(fldr, "%d.png"%(y,))
          try:
            os.makedirs(fldr)
          except:
            pass  # war wohl schon existent == macht nix
          try:
            img.save(fl)
          except:
            if DEBUG: print "Error img.save()\n",
            pass
      self.queue.task_done()
      self.__decRunningCount()


  # ######################################################################
  # Der += bzw. -= scheint nicht "atomar" implementiert zu sein - daher
  # muss der Schreibzugriff per Lock() Thread-safe gemacht werden.
  def __incRunningCount(self):
    self.running_lock.acquire()
    self.running_count+=1
    self.running_lock.release()
  def __decRunningCount(self):
    self.running_lock.acquire()
    self.running_count-=1
    self.running_lock.release()


  # ######################################################################
  # Das OSM-Tile(x, y, z) aus dem Web laden und samt Status zurückliefern.
  def __getTileFromWeb(self, x, y, z):
    if DEBUG: print "Web-request(%d,%d,%d)\n"%(x, y, z),
    url=self.tile_url.format(z=z, x=x, y=y)
    try:
      response=requests.get(url, timeout=self.request_timeout)
    except:
      if DEBUG: print "Timeout(%d,%d,%d)\n"%(x, y, z),
      return((self.status["FAILED"], None))
    try:
      img=Image.open(BytesIO(response.content))
    except:
      if DEBUG: print "Error in Image.open()\n",
      return((self.status["FAILED"], None))
    return((self.status["GOOD"], img))



# ######################################################################
# Ein kleiner Cache im RAM.
# Ganz ohne Cache kann OSMscr2 nicht arbeiten, weil mögliche
# Tile-Nachlieferungen ja irgendwo hin müssen.
# Für einen lokalen Tile-Server könnte natürlich auch komplett auf
# synchrones Laden der Tiles gewechselt werden.... also quasi so, dass
# getTile() direkt und nur __getTileFromWeb() aufruft.
class RAMcache():
  def __init__(self):
    self.ram_cache=dict()
    self.max_tiles=CACHE_MAX_TILES
    self.age=0
    self.oldest=0
    self.lock=threading.Lock()  # Schreibzugriffe auf self.ram_cache serialisieren


  # ######################################################################
  # Legt das Tile(x, y, z) im RAM ab und löscht bei Erreichen der
  # Tile-Maximal-Anzahl die Tiles, die am längsten nicht angefasst wurden.
  def putIntoRAM(self, x, y, z, tile):
    self.lock.acquire()
    lv=len(self.ram_cache)
    if len(self.ram_cache)>self.max_tiles:
      del_cnt=(self.age-self.oldest)/10     # die ältesten 10% löschen - naja grob 10%
      self.oldest+=del_cnt
      del_lst=list()
      for xyz, (dummy, age) in self.ram_cache.items():
        if age<self.oldest:
          del_lst.append(xyz)
      for xyz in del_lst:
        del self.ram_cache[xyz]
    self.ram_cache.update({(x, y, z):(tile, self.age)})
    self.age+=1
#    print "len:%3d-%3d  age:%5d  oldest:%5d\n"%(lv, len(self.ram_cache), self.age, self.oldest),
    self.lock.release()


  # ######################################################################
  # Liefert das Tile(x, y, z) aus dem RAM - oder None, wenn es dort noch
  # nicht existiert.
  def getFromRAM(self, x, y, z):
    tile, age=self.ram_cache.get((x, y, z), (None, None))
    if tile is not None:
      self.lock.acquire()
      self.ram_cache.update({(x, y, z):(tile, self.age)})
      self.age+=1
      self.lock.release()
    return(tile)
