#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ###########################################################
# Ein kleines GUI als Test der OSM_screen-Klasse.
# Also Anzeige der OpenStreetMap-Karte mit der Möglichkeit
# des scrollens, zoomens und einer Track-Anzeige.
#
# Detlev Ahlgrimm, 2017
#
#
# 08.04.2017  erste Version
# 10.04.2017  kompletter Umbau
# 14.04.2017  scrollMapToPoint() und Server-Wahl mittels SOURCE eingebaut
# 15.04.2017  Zoomlevel-Anpassung nach Track-Laden eingebaut
# 16.04.2017  Umbau wegen Umstellung auf wx.MemoryDC(), Auslagerung von Funktionen
# 17.04.2017  unabhängige Point-Listen und drawLinesForAP() zugefügt
#             Entfernungs-Anzeige bei AP-Sichtungs-Punkten
# 23.04.2017  einiges nach OSMhelper ausgelagert, getPausePoints() zugefügt
# 30.04.2017  Slider-Frame samt Verarbeitung von dessen Daten eingebaut
# 02.05.2017  Slider-Frame um Autofollow erweitert
# 04.05.2017  mittlerweile unnötige Funktionen gelöscht
# 15.05.2017  Menüfunktion "remove track/all tracks" eingebaut
# 18.05.2017  Menüfunktion "hilite track" eingebaut, Track-Länge zugefügt,
#             Abfrage von Ctrl bei FileDrop für "alte Tracks löschen"
# 03.06.2017  Umstellung auf OSMscr3
# 20.06.2017  Umstellung auf TrackData()
# 21.06.2017  SliderFrame umgebaut, so dass der Frame ohne Timer läuft
# 22.06.2017  in Map-verändernden Funktionen je ein finishAutofollow() zugefügt
#             (um dadurch zu umgehen, dass die Funktion auf beide Maps wirken müsste).

import wx               # python-wxWidgets-2.8.12.1-10.4.1.x86_64
import time
import math
import os
from datetime import datetime, tzinfo, timedelta
from time import mktime
from calendar import timegm
import requests
from Queue import Queue

from OSMscr3 import OSMscr3
from OSMhelper import OSMhelper

try:
  from Database4wd        import Database
  DATABASE_FILE="/home/dede/daten/wardriving.sqlite"
  wardriving_db_found=True
except:
  wardriving_db_found=False

VERSION="1.4c"

HOME_DIR=os.path.expanduser('~')
# http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Tile_servers
# http://www.thunderforest.com/docs/apikeys/
SOURCE={
  "OSM"   : { "tile_url"  : "http://tile.openstreetmap.org/{z}/{x}/{y}.png",
              "tile_size" : 256,
              "max_zoom"  : 19,
              "cache_dir" : os.path.join(HOME_DIR, "wd", "osm")
            },
  "OSMnc" : { "tile_url"  : "http://tile.openstreetmap.org/{z}/{x}/{y}.png",
              "tile_size" : 256,
              "max_zoom"  : 19,
              "cache_dir" : None    # RAM-Cache
            },
  "OCM"   : { "tile_url"  : "https://tile.thunderforest.com/cycle/{z}/{x}/{y}.png?apikey=0ab....selber.einen.holen...9625",
              "tile_size" : 256,
              "max_zoom"  : 18,
              "cache_dir" : os.path.join(HOME_DIR, "wd", "ocm")
            },
  "FAU"   : { "tile_url"  : "https://osm.rrze.fau.de/osmde/{z}/{x}/{y}.png",
              "tile_size" : 256,
              "max_zoom"  : 19,
              "cache_dir" : os.path.join(HOME_DIR, "wd", "fau")
            },
  "FAUHD" : { "tile_url"  : "https://osm.rrze.fau.de/osmhd/{z}/{x}/{y}.png",
              "tile_size" : 512,
              "max_zoom"  : 19,
              "cache_dir" : os.path.join(HOME_DIR, "wd", "fauhd")
            },
}
SERVER="OSM"

DEBUG=True

# ###########################################################
# eine kleine Klasse, um neue Dateien via Drag&Drop öffnen
# zu können.
class FileDrop(wx.FileDropTarget):
  def __init__(self, window):
    wx.FileDropTarget.__init__(self)
    self.window=window
    self.Ctrl_pressed=False

  def OnDropFiles(self, x, y, filenames):
    self.window.fileDropped(filenames)

  def OnDragOver(self, x, y, d):
    if d==wx.DragCopy:
      if wx.GetKeyState(wx.WXK_CONTROL):
        self.Ctrl_pressed=True
        return(wx.DragLink)
      self.Ctrl_pressed=False
      return(wx.DragCopy)
    return(d)


# ######################################################################
# Eine Datenstruktur für die Attribute eines Tracks.
class TrackData():
  def __init__(self):
    self.start_ts=None
    self.end_ts=None
    self.pause_points_handle=None
    self.track=None
    self.dist=0.0
    self.filename=None

# ######################################################################
# Eine Datenstruktur für die Variablen beim Autofollow.
class AutofollowData():
  def __init__(self):
    self.track_handle=None            # das Handle des nachzufahrenden Tracks
    self.next_pos=None                # die nächste Zielposition als Index (index, sub_index)
    self.next_posLL=None              # die nächste Zielposition als Geo-Koordinate
    self.cur_points=[]                # der aktuelle Abschnitt bis zum NewMapPoint
    self.cur_index=0                  # der Index in cur_points
    self.step_size=1                  # die Anzahl der Pixel pro Schritt 
    self.cur_zoom=None                # der Zoom-Level, für den cur_points gültig ist
    self.OSMscr_handles=(None, None)  # die zwei Handles zur asychronen Map-Vorberechnung
    self.cur_map_handle=0             # der Index in OSMscr_handles (==aktuell angezeigte Map)


# ######################################################################
# Das eigentliche Fenster mit der OSM-Ansicht.
class MapWindow(wx.Window):
  def __init__(self, parent, centerLL, zoom, border, title, size):
    wx.Window.__init__(self, parent)
    self.parent=parent      # der Frame
    self.centerLL=centerLL  # die lat/lon-Koordinate in der Mitte des Fensters
    self.zoom=zoom          # der aktuelle Zoom-Level der OSM-Tiles
    self.border=border      # die Anzahl der unsichbaren Tiles pro Seite (horizontal, vertikal)

    self.parent.Bind(wx.EVT_CLOSE,  self.onClose)
    self.Bind(wx.EVT_PAINT,         self.onPaint)
    self.Bind(wx.EVT_LEFT_DOWN,     self.onLeftDown)
    self.Bind(wx.EVT_LEFT_DCLICK,   self.onLeftDClick)
    self.Bind(wx.EVT_RIGHT_DOWN,    self.onRightDown)
    self.Bind(wx.EVT_LEFT_UP,       self.onLeftUp)
    self.Bind(wx.EVT_MOTION,        self.onMotion)
    self.Bind(wx.EVT_MOUSEWHEEL,    self.onMouseWheel)
    self.Bind(wx.EVT_SIZE,          self.onSize)
    self.Bind(wx.EVT_KEY_DOWN,      self.onKeyDown)
    self.Bind(wx.EVT_KEY_UP,        self.onKeyUp)
    self.Bind(wx.EVT_CONTEXT_MENU,  self.onContextMenu)

    self.timer_refresh_interval=500
    self.timer_refresh=wx.Timer(self)                                 # der Timer zum Nachladen von fehlenden Tiles
    self.Bind(wx.EVT_TIMER, self.onTimerRefresh, self.timer_refresh)
    self.timer_refresh.Start(self.timer_refresh_interval)

    self.timer_slider_interval=100
    self.timer_slider=wx.Timer(self)                                  # der Timer zur Kommunikation mit dem SliderFrame
    self.Bind(wx.EVT_TIMER, self.onTimerSlider, self.timer_slider)
    self.timer_slider.Start(self.timer_slider_interval)

    self.timer_autofollow_interval=10
    self.timer_autofollow=wx.Timer(self)                              # der Timer zur Kommunikation mit dem SliderFrame
    self.Bind(wx.EVT_TIMER, self.onTimerAutofollow, self.timer_autofollow)

    self.statusbar=self.parent.CreateStatusBar(5)
    self.statusbar.SetStatusWidths([50, 50, 100, -1, -1])
    self.dc_sizeXY=self.parent.GetSize() # das ist schon die Größe abzüglich der Statusbar

    self.hlp=OSMhelper()
    self.osm=OSMscr3(SOURCE[SERVER])
    self.handle=self.osm.setup(self.dc_sizeXY, centerLL, self.zoom, self.border)

    self.wardriving_mode=False          # wird per Kontext-Menü getoggled und steuert die Anzeige der AP-Koordinaten-Punkte
    self.wardriving_point_handle=None   # das Handle für die angezeigte AP-Koordinaten-Punkte-Liste
    self.handle_for_ap_list=None        # ein Tupel mit zwei Handles der angezeigten Sichtungs-Punkte-Listen (norm, high)
    self.ap_seen_at_list=None           # Tupel aus AP-Koordinate und Sichtungs-Punkte-Liste zur Darstellung der Linien
    self.trackdata=dict()               # die Daten der geladenen Tracks, Format: {track_handle : TrackData()}
    self.leftDown=False                 # entspricht dem Up/Down-Zustand der linken Maus-Taste
    self.esc=False                      # entspricht dem Up/Down-Zustand der ESC-Taste
    self.ctrlIsDown=False               # entspricht dem Up/Down-Zustand der Strg-Taste
    self.shiftIsDown=False              # entspricht dem Up/Down-Zustand der Shift-Taste
    self.refresh_needed=False           # auf True, wenn noch Leer-Tiles in der Map enthalten sind
    self.was_moved=False                # auf True, wenn bei leftDown Maus-Cursor-Bewegung stattfand
    self.infoText=None                  # der Text für ein Balloon-Popup nach Doppelklick
    self.slider_frame=None              # der Slider-Frame
    self.slider_queue=None              # die Rückgabe-Queue des Slider-Frames
    self.slider_track_handle=None       # das im SliderFrame aktuell geladene Track-Handle
    self.autofollow_is_running=False    # auf True, wenn Autofollow eingeschaltet ist
    self.af=AutofollowData()            # diverse Variablen für Autofollow
    self.on_track=False                 # auf True, wenn Autofollow "ungestört" läuft, Map-Verschiebungen setzen auf False
    self.hilite_track_handle=-1         # das Handle des Tracks, der gerade hervorgehoben ist
    self.last_painted_handle=-1         # das in onPaint() zuletzt dargestellte self.handle
    self.newMapPoints_handle=None       # das Point-Handle für NewMapPoints, die in die Map geladen wurden
    self.track_point_data=None          # Geschwindigkeit, Entfernung und Zeitdauer des aktuellen Trackpoints für onPaint()
    self.menu_show_grid=False           # Status des Menüpunkts
    self.menu_show_track_points=False   # Status des Menüpunkts
    self.menu_show_new_map_points=False # Status des Menüpunkts
    self.debug_running=False            # ein Flag zur Blockade von debug(), wenn es bereits läuft

    self.statusbar.SetStatusText("%d"%(self.zoom,), 0)

    self.filename_dt=FileDrop(self)
    self.SetDropTarget(self.filename_dt)

    if wardriving_db_found:
      self.db=Database(DATABASE_FILE)
    else:
      self.point_list=[ ( (54.79167946079812, 9.442416429519653), 5,  (255, 0, 0),      (99, " Hackspace ")),
                        ( (54.80507860419667, 9.52488899230957),  8,  (255, 215, 0),    (99, " Zuhause\n ist es am schönsten ")),
                        ( (54.80737577256639, 9.513548612594604), 5,  (255, 0, 0),      (99, " lecker\n Bier-Versorger ")),
                      ]
      self.osm.loadPoints(self.point_list)
    self.pointMoveOver=list()     # wird in onMotion auf die Points-Daten unter dem Maus-Cursor gesetzt

    self.osm.setShowGrid(self.menu_show_grid)
    wx.FutureCall(10, self.SetFocus)  # damit EVT_KEY_xx kommt


  # ###########################################################
  # Wird aus FileDrop aufgerufen, wenn ein oder mehrere
  # DateiObjekte auf dem Fenster "fallengelassen" wurden.
  # Jede Datei kann einen oder mehrere Tracks oder 
  # Tracksegmente enthalten.
  def fileDropped(self, filenames):
    all_lstsLL=list()
    if self.filename_dt.Ctrl_pressed:                 # wenn Strg-Taste gedrückt war (sollen bisherige Tracks gelöscht werden)
      if len(self.trackdata)>0:                       # wenn bereits Tracks angezeigt werden
        for handle in self.trackdata:
          self.osm.unloadTrack(handle)                                      # Tracks löschen
          self.osm.unloadPoints(self.trackdata[handle].pause_points_handle) # und ihre Pause-Punkte auch
        self.osm.forceRefresh()                       # in jedem Fall die Anzeige neu machen (falls die neuen Tracks illegal sind)
        self.trackdata=dict()
        self.closeSliderFrame()                       # und noch den Slider weg
    for fn in filenames:                              # über alle neuen Track-Dateinamen
      if self.isTrackAlreadyLoaded(fn):               # wenn Track schon geladen ist
        print "File already loaded:", fn
        continue                                      # Track überspringen
      lstsLL, wptsLL=self.hlp.loadGPX(fn)             # laden (als Liste von Tracks bzw. Track-Segmenten)
      for trk in lstsLL:
        all_lstsLL.append((trk, wptsLL, fn))          # und die Track-Segmente der Gesamt-Liste zufügen
        wptsLL=list()                                 # wpts jeweils nur zum ersten Track-Segment in die Map schreiben

    # Format von all_lstsLL:
    #   [ ([((lat, lon), time), ((lat, lon), time), ...], [...waypoints...]),   - erster Track + Timestamps
    #     ([((lat, lon), time), ((lat, lon), time), ...], [...waypoints...]),   - zweiter Track + Timestamps
    #      ... ]
    trk=None
    no_tracks_loaded=len(self.trackdata)==0
    for trk, wptsLL, fn in all_lstsLL:
      track_handle=self.osm.loadTrack(trk, width=2)
      pause_points_handle=self.addPausePoints(trk, wptsLL, track_handle)
      data=TrackData()
      data.start_ts=trk[0][1]
      data.end_ts=trk[-1][1]
      data.pause_points_handle=pause_points_handle
      data.track=trk
      data.dist=self.hlp.getTrackLength(trk)
      data.filename=fn
      self.trackdata.update({track_handle:data})
    if len(all_lstsLL)>0:                           # wenn neue Tracks geladen wurden
      self.finishAutofollow()                       # dann ggf. Autofollow stoppen (wegen des zweiten OSMscr3-Handles)
      self.showSliderFrame()                        # SliderFrame öffnen
      if self.slider_track_handle is None:          # wenn noch kein Track im Slider geladen ist
        th=min(self.trackdata)                      # den Track mit dem kleinsten Handle...
        self.loadTrackIntoSlider(th)                # ...in den Slider laden
      if no_tracks_loaded:                          # wenn bisher keine Tracks geladen waren
        self.setZoomAndCenterForTracks(all_lstsLL)  # Map so einstellen, dass alle Tracks sichtbar sind
    else:
      print "track is illegal!"
    self.osm.forceRefresh()
    self.Refresh()


  # ###########################################################
  # Liefert True, wenn "filename" bereits in einem geladenen
  # Track vorkommt.
  def isTrackAlreadyLoaded(self, filename):
    for handle, data in self.trackdata.items():
      if data.filename==filename:
        return(True)
    return(False)


  # ###########################################################
  # Setzt für allen Pausen-Punkte in "trk" und alle Punkte in
  # "wpts" einen Point in der self.osm-Map. 
  # Zurückgeliefert wird das Handle, mit dem die gesetzten
  # Punkte wieder gelöscht werden können.
  def addPausePoints(self, trk, wpts, track_handle_debug_only):
    if len(trk)==0 and len(wpts)==0:
      return(None)
    pl=list()
    for wpt, data in wpts:
      txt=""
      if "name" in data:  txt+=" name: %s \n"%(data["name"],)
      if "desc" in data:  txt+=" desc: %s \n"%(data["desc"],)
      if "type" in data:  txt+=" type: %s \n"%(data["type"],)
      pl.append((wpt, 5, (255, 0, 0), (99, txt[:-1])))
    if len(trk)>0:
      ppt_lst=self.hlp.getPausePoints(trk)
      for ppt in ppt_lst:
        mins=(ppt[3]-ppt[2])/60
        secs=(ppt[3]-ppt[2])-mins*60
        data=(self.hlp.getUTCTimestampAsLocal(ppt[2]), self.hlp.getUTCTimestampAsLocal(ppt[3]), "%d:%02d"%(mins, secs))
        if mins<5:
          pl.append(((ppt[0], ppt[1]), 5, (255, 255, 0), (2, data)))   # wohl nur Pinkel-Pause
        else:
          pl.append(((ppt[0], ppt[1]), 5, (0, 128, 0), (2, data)))
      pl.append((trk[0][0], 5, (255, 0, 0), (99, " Start ")))
      pl.append((trk[-1][0], 5, (255, 0, 0), (99, " Destination ")))

    if self.menu_show_track_points:
      for idx in range(len(trk)):
        tp, ts=trk[idx]
        speed, dist, secs=self.osm.all_tracks[track_handle_debug_only].point_data[idx]
        pl.append((tp, 2, (128, 0, 0), (99, " %d:%6.2f km/h, %6.2f km, %s "%(idx, speed, dist, self.secsToHMS(secs)))))

    if len(pl)>0:
      pause_points_handle=self.osm.loadPoints(pl)
    else:
      pause_points_handle=None
    return(pause_points_handle)


  # ######################################################################
  # Liefert für eine Anzahl von Sekunden in "secs" dessen entsprechenden
  # Wert im Format HH:MM:SS als String.
  # https://stackoverflow.com/a/775075/3588613
  def secsToHMS(self, secs):
    m, s=divmod(secs, 60)
    h, m=divmod(m, 60)
    return("%d:%02d:%02d"%(h, m, s))


  # ######################################################################
  # Den Ausschnitt so anpassen, dass "lstsLL" komplett sichtbar wird.
  def setZoomAndCenterForTracks(self, lstsLL):
    t0LL=list()
    for trk, dummy, dummy in lstsLL:    # Liste von Listen in eine flache...
      for pLL, dummy in trk:            # ...Liste umwandeln
        t0LL.append(pLL)
    zoom, centerLL=self.osm.getZoomAndCenterForAllPointsVisible(t0LL)
    self.setZoom(zoom)
    self.osm.setCenter(centerLL)


  # ######################################################################
  # Wenn das Hauptfenster geschlossen werden soll, muss vorher ein ggf.
  # offenes Slider-Fenster weg.
  def onClose(self, evt):
    if self.slider_frame is not None:
      self.closeSliderFrame()
    # das Hauptfenster muss etwas verzögert geschlossen werden, damit das
    # Slider-Fenster noch seinen "CLOSE" auf die Queue schreiben kann.
    wx.FutureCall(10, self.parent.Destroy)


  # ######################################################################
  # Fensterinhalt darstellen.
  def onPaint(self, evt):
    new_bmp, (sx, sy)=self.osm.drawMap(self.handle)
    if new_bmp or self.last_painted_handle!=self.handle:
      self.last_painted_handle=self.handle
      (w, h, bmpdata), (sx, sy), self.refresh_needed=self.osm.getBitmap(self.handle)
      self.bmpbuf=wx.BitmapFromBuffer(w, h, bmpdata)
    self.dc=wx.PaintDC(self)
    self.dc.DrawBitmap(self.bmpbuf, sx, sy)

    # AP-Sichtungs-Punkte mittels Linien mit der AP-Koordinate verbinden
    if self.ap_seen_at_list is not None:
      self.drawLinesForAP()

    # Geschwindigkeit anzeigen
    if self.slider_frame is not None and self.on_track:
      self.drawFollowPoint((self.dc_sizeXY[0]/2, self.dc_sizeXY[1]/2))
      if self.track_point_data is not None:
        txt=" %.2f km/h \n %.2f km \n %s "%(self.track_point_data[0], self.track_point_data[1], self.secsToHMS(self.track_point_data[2]))
        self.showBalloonText((self.dc_sizeXY[0]/2, self.dc_sizeXY[1]/2), txt)

    # Maus-über-Point-Popup anzeigen
    if len(self.pointMoveOver)>0:
      balloon_lst=list()
      for pos, (ptype, data) in self.pointMoveOver:
        if ptype==0:    # Punkt-Typ == AP-Koordinate
          txt=" %s \n %s \n channel: %d \n %s \n points: %d \n %s \n %s "%data[:7]
        elif ptype==1:  # Punkt-Typ == AP-Sichtungs-Koordinate
          txt=" dbm:%d \n %s \n distance: %dm "%data
        elif ptype==2:  # Punkt-Typ == Pausen-Punkt
          txt=" *** pause ***\n start:%s \n dest:%s \n duration: %s Min."%data
        elif ptype==99: # Testdaten bzw. einzelner String
          txt=data
        else:
          txt=" ??? "
        balloon_lst.append((pos, txt))
      if len(balloon_lst)>0:
        self.showBalloonTextList(balloon_lst)

    # Info-Popup anzeigen
    if self.infoText is not None:
      self.showBalloonText(self.infoText[0], self.infoText[1])
      self.infoText=None

    self.statusbar.SetStatusText(str(self.osm.getOpenRequests()), 1)


  # ######################################################################
  # Stellt für die Track-Anzeige den Startpunkt und bei Auto-Follow die
  # aktuelle Positionsmarkierung dar.
  def drawFollowPoint(self, posXY):
    self.dc.SetPen(wx.Pen("BLACK"))
    self.dc.SetBrush(wx.Brush("WHITE"))
    self.dc.DrawCirclePoint(posXY, 5)


  # ######################################################################
  # Zeichnet Linien von einer AP-Position zu allen seinen
  # Sichtungs-Punkten.
  def drawLinesForAP(self):
    base, lst=self.ap_seen_at_list
    shift=self.osm.getShift()
    self.dc.SetPen(wx.Pen("BLACK"))
    for pLL, xx, xx, (xx, (dbm, ts, dist)) in lst:
      x, y=self.osm.getPixelForLatLon(pLL)
      self.dc.DrawLine(base[0]-shift[0], base[1]-shift[1], x, y)


  # ######################################################################
  # Linke Maustaste geklickt - Map scrollen.
  def onLeftDown(self, evt):
    self.leftDown=True
    self.was_moved=False
    self.leftDownPosXY=evt.GetPosition()
    self.SetCursor(wx.StockCursor(wx.CURSOR_HAND))
    self.statusbar.SetStatusText("%dx%d"%(self.leftDownPosXY[0], self.leftDownPosXY[1]), 2)
    posLL=self.osm.getLatLonForPixel(self.leftDownPosXY)
    self.statusbar.SetStatusText("%14.12f, %14.12f"%(posLL[0], posLL[1]), 3)

    self.pointMoveOver=self.osm.getDataForPoint(self.leftDownPosXY)
    # ggf. eine alte Sichtungs-Liste entfernen
    if self.handle_for_ap_list is not None:
      self.removeAPSeenAtList()
      if len(self.pointMoveOver)==0:  # wenn nicht eh gleich ein Refresh erfolgt
        self.osm.forceRefresh()
        self.Refresh()                # hier anfordern

    # ggf. eine neue Sichtungs-Liste erstellen und in die Map laden
    if wardriving_db_found and len(self.pointMoveOver)>0:
      for pos, (ptype, data) in self.pointMoveOver: # wenn eine neue Sichtungs-Liste angezeigt werden soll
        if ptype==0:  # wenn AP-Koordinate
          # data=(bssid, name, channel, crypto, calc_points, first_seen, last_seen, lat, lon)
          bssid=data[0]
          locs=self.db.getFullGPSpoints(bssid)  # locs=[(lat, lon, dbm), ...]
          max_dbm=-200
          for lat, lon, dbm, ts in locs:  # größten DBM-Wert suchen, um ihn farblich hervorheben zu können
            max_dbm=max(max_dbm, int(dbm))
          norm_lst=list()   # die Liste für die Points mit den nicht-größten DBM-Werten
          high_lst=list()   # die Liste für die Points mit den größten DBM-Werten
          for lat, lon, dbm, ts in locs:
            dist=int(round(self.hlp.haversine((data[7], data[8]), (float(lat), float(lon)))*1000))
            if int(dbm)==max_dbm:
              high_lst.append(((float(lat), float(lon)), 3, (255, 255, 0), (1, (int(dbm), ts, dist))))
            else:
              norm_lst.append(((float(lat), float(lon)), 3, (0, 0, 255), (1, (int(dbm), ts, dist))))
          self.ap_seen_at_list=(pos, norm_lst+high_lst)
          norm=self.osm.loadPoints(norm_lst)  # erst die blauen Punkte
          high=self.osm.loadPoints(high_lst)  # und dann die gelben drüber (OSMscr3 sortiert nach Handle)
          self.handle_for_ap_list=(norm, high)
          self.osm.forceRefresh()
          self.Refresh()
          break # einer langt - sonst wirds unübersichtlich


  # ######################################################################
  # Linke Maustaste doppel-geklickt - Infodaten anzeigen.
  def onLeftDClick(self, evt):
    doubleClickPosXY=evt.GetPosition()
    posLL=self.osm.getLatLonForPixel(doubleClickPosXY)
    txt=self.hlp.getNominatim(posLL, self.zoom)
    self.infoText=(doubleClickPosXY, " %s "%(txt,))
    self.Refresh()


  # ######################################################################
  # Rechte Maustaste geklickt - nur Klick-Position merken.
  def onRightDown(self, evt):
    self.rightDownPosXY=evt.GetPosition()
    evt.Skip(True)  # weiter zu wx.EVT_CONTEXT_MENU


  # ######################################################################
  # Linke Maustaste losgelassen - Map scrollen beenden.
  def onLeftUp(self, evt):
    self.leftDown=False
    self.SetCursor(wx.StockCursor(wx.CURSOR_ARROW))
    if self.was_moved:
      if self.ap_seen_at_list is not None:  # wenn gerade AP-Sichtungspunkte angezeigt werden
        base, lst=self.ap_seen_at_list
        shift=self.osm.getShift()
        base=(base[0]-shift[0], base[1]-shift[1])
        self.ap_seen_at_list=(base, lst)    # dann die XY-Position der AP-Koordinate korrigieren
      self.was_moved=False
      self.osm.endMoveMap()
      self.Refresh()
      # der Refresh() wird hier gebraucht, um nach einem Verschiebevorgang
      # außerhalb der vorgeladenen Rand-Tiles die fehlenden Tiles nachzuladen.


  # ######################################################################
  # Mauscursor wird über das Fenster bewegt.
  def onMotion(self, evt):
    self.cur_mouse_posXY=evt.GetPosition()
    if self.leftDown:                   # wenn dabei die linke Maustaste gedrückt gehalten wird
      self.was_moved=True
      self.on_track=False
      scroll_distXY=self.leftDownPosXY-self.cur_mouse_posXY
      self.osm.doMoveMap(scroll_distXY) # Map verschieben
      self.Refresh()
    else:                               # Maustaste nicht gedrückt -> ggf. Sprechblase anzeigen
      old_len=len(self.pointMoveOver)
      self.pointMoveOver=self.osm.getDataForPoint(self.cur_mouse_posXY)
      if len(self.pointMoveOver)>0:     # wenn Daten verfügbar sind
        self.Refresh()                  # Sprechblase darstellen
      elif old_len>0:                   # wenn vorher Daten verfügbar waren
        self.Refresh()                  # die Sprechblase wieder löschen


  # ######################################################################
  # Gibt eine Sprechblase mit "text" bei der Position "(x, y)" aus.
  def showBalloonText(self, (x, y), text):
    self.dc.SetPen(wx.Pen("BLACK"))
    self.dc.SetBrush(wx.Brush("WHITE"))
    w, h=self.dc.GetTextExtent(text)
    d=5                 # Abstand der Sprechblase zur Position (x, y)
    xo, yo=(x+d, y-d-h)
    if yo<0: yo=y+d
    if xo+w>self.dc_sizeXY[0]:  xo=x-d-w
    self.dc.DrawRectangle(xo, yo, w, h)
    self.dc.DrawText(text, xo, yo)


  # ######################################################################
  # Wie showBalloonText() - nur für mehrere nebeneinander.
  def showBalloonTextList(self, lst):
    maxcnt=5
    wsum=0    # Gesamtbreite
    xlst=[0]  # Startpositionen
    wlst=[]   # Breiten
    maxh=0    # Höhe

    for pos, text in lst:
      w, h=self.dc.GetTextExtent(text)
      wlst.append(w)
      wsum+=w
      xlst.append(wsum)
      maxh=max(maxh, h)
    self.dc.SetPen(wx.Pen("BLACK"))
    self.dc.SetBrush(wx.Brush("WHITE"))
    x, y=lst[0][0]
    d=5 # Abstand
    xo, yo=(x+d, y-d-maxh)
    if yo<0: yo=y+d
    if xo+wsum>self.dc_sizeXY[0]:  xo=x-d-wsum
    if xo<0:  xo=0
    idx=0
    for pos, text in lst:
      self.dc.DrawRectangle(xo+xlst[idx], yo, wlst[idx], maxh)
      self.dc.DrawText(text, xo+xlst[idx], yo)
      idx+=1


  # ######################################################################
  # Scrollrad betätigt -> Map ändern.
  def onMouseWheel(self, evt):
    if evt.GetWheelRotation()>0:
      zoom_new=min(SOURCE[SERVER]["max_zoom"], self.zoom+1)
    else:
      zoom_new=max(0, self.zoom-1)
    if zoom_new!=self.zoom:
      self.setZoom(zoom_new, evt.GetPosition())


  # ######################################################################
  # Ändert den Zoom-Level.
  def setZoom(self, zoom_new, pos=(0, 0)):
    if self.zoom!=zoom_new:                                   # Zoom-Level wurde verändert
      self.zoom=zoom_new
      self.removeAPSeenAtList()                               # ggf. eine alte Sichtungs-Liste entfernen
      self.on_track=False
      self.osm.setZoom(self.zoom, pos)                        # neuen Zoom-Level für Map einstellen
      self.statusbar.SetStatusText("%d"%(self.zoom,), 0)

      if self.menu_show_new_map_points:                       # wenn NewMapPoints dargestellt werden sollen
        if len(self.trackdata)==0:                            # wenn kein Track [mehr] geladen ist
          if self.newMapPoints_handle is not None:            # wenn gerade NewMapPoints dargestellt werden
            self.osm.unloadPoints(self.newMapPoints_handle)   # diese aus der Map löschen
            self.newMapPoints_handle=None                     # und auch hier als gelöscht markieren
        else:                                                 # wenn ein Track geladen ist
          if self.slider_track_handle is not None:            # wenn ein Track im Slider geladen ist
            pl=self.osm.getNewMapPointsForTrack(self.slider_track_handle) # die NewMapPoints neu berechnen
            pts=list()
            for idx, lst in sorted(pl.items()):               # eine ladbare Point-Liste daraus bauen
              for tc, p in lst:
                pts.append((p, 5, (0, 0, 255), (99, " %d,%d "%(idx, tc))))
            if self.newMapPoints_handle is not None:          # wenn noch alte NewMapPoints dargestellt werden
              self.osm.unloadPoints(self.newMapPoints_handle) # weg damit
            self.newMapPoints_handle=self.osm.loadPoints(pts) # und die neuen NewMapPoints laden
      self.Refresh()


  # ######################################################################
  # Wird bei Betätigung der Taste "D" aufgerufen.
  def debug(self):
    if self.debug_running:
      return
    self.debug_running=True


    self.debug_running=False
    self.Refresh()


  # ######################################################################
  # Fenstergröße wurde geändert - Variablen anpassen.
  def onSize(self, evt):
    self.dc_sizeXY=self.GetSize()
    try:
      self.on_track=False
      self.osm.setSize(self.dc_sizeXY)
    except:
      pass  # der allererste EVT_SIZE kommt gerne mal, bevor self.osm existiert


  # ######################################################################
  # Eine Taste wurde betätigt.
  def onKeyDown(self, evt):
    if evt.GetKeyCode()==wx.WXK_CONTROL:    self.ctrlIsDown=True
    if evt.GetKeyCode()==wx.WXK_SHIFT:      self.shiftIsDown=True
    if evt.GetKeyCode()==wx.WXK_ESCAPE:     self.esc=True
    evt.Skip(True)


  # ######################################################################
  # Eine Taste wurde losgelassen.
  def onKeyUp(self, evt):
    if evt.GetKeyCode()==wx.WXK_CONTROL:  self.ctrlIsDown=False
    if evt.GetKeyCode()==wx.WXK_SHIFT:    self.shiftIsDown=False
    if evt.GetKeyCode()==wx.WXK_ESCAPE:   self.esc=False
    if evt.GetKeyCode()==ord("H"):
      self.osm.setCenter((54.805109, 9.524913))
      self.on_track=False
      self.Refresh()
    if evt.GetKeyCode()==ord("S"):        self.showSliderFrame()
    if evt.GetKeyCode()==ord("D"):        self.debug()
    evt.Skip(True)


  # ###########################################################
  # Ruft einmal pro Update-Intervall die Prüfung wegen Refresh
  # auf, um damit ggf. mittlerweile verfügbare Tiles anzuzeigen.
  def onTimerRefresh(self, evt):
    if self.refresh_needed:
      self.Refresh()


  # ###########################################################
  # Ruft die Update-Funktion für den SliderFrame auf sofern
  # der SliderFrame offen ist.
  def onTimerSlider(self, evt):
    if self.slider_frame is not None:                         # wenn der SliderFrame überhaupt offen ist
      while self.slider_queue.qsize()>0:                      # wenn vom SliderFrame irgendwas gemeldet wurde
        task, value=self.slider_queue.get()
        if task=="CLOSE":                                     # wenn Slider-Frame meldet, dass er manuell beendet wurde
          self.finishAutofollow()                             # dann ggf. Autofollow stoppen
          self.closeSliderFrame(is_already_closed=True)       # und den Slider dann auch hier löschen
          break # die Queue ist jetzt weg - daher muss die while-Schleife verlassen werden
        elif task=="SCROLL":                                  # wenn die Slider-Position manuell verändert wurde
          if not self.autofollow_is_running:                  # wenn Autofollow nicht läuft und
            if self.slider_track_handle is not None:          # wenn ein Track im Slider geladen ist
              posLL=self.trackdata[self.slider_track_handle].track[value]
              self.osm.setCenter(posLL[0], self.handle)       # den Map-Mittelpunkt gemäß Slider-Positionsänderung ausrichten
              self.track_point_data=self.osm.getTrackPointData(self.slider_track_handle, value)
              self.on_track=True
              self.Refresh()
        elif task=="FREQ":                                    # wenn die Frequenz geändert wurde
          if self.timer_autofollow_interval!=value:           # und auch tatsächlich anders als zuvor ist
            self.timer_autofollow_interval=value              # Änderung festschreiben
            if self.autofollow_is_running:                    # wenn gerade Autofollow läuft
              self.timer_autofollow.Stop()                    # direkt den Timer entsprechend anpassen
              self.timer_autofollow.Start(self.timer_autofollow_interval)
        elif task=="STEP":                                    # wenn die Schrittweite geändert wurde
          if self.autofollow_is_running:                      # wenn gerade Autofollow läuft
            self.af.step_size=value                           # Änderung festschreiben
        elif task=="AUTO":                                    # wenn Autofollow bzgl. ein/aus umgeschaltet wurde
          if value==1:                                        # wenn Autofollow eingeschaltet wurde
            self.prepareAutofollow()                          # Autofollow starten
            self.autofollow_is_running=True
            self.timer_autofollow.Start(self.timer_autofollow_interval)
          else:                                               # wenn Autofollow ausgeschaltet wurde
            self.finishAutofollow()                           # Autofollow beenden


  # ###########################################################
  # Initialisiert die Variablen für Autofollow.
  def prepareAutofollow(self):
    if self.slider_track_handle is not None:                    # nur wenn ein Track im Slider geladen ist
      self.af.track_handle=self.slider_track_handle             # Handle übertragen
      self.af.OSMscr_handles=(self.handle, self.osm.newMap())   # zweite Map anlegen
      self.af.cur_map_handle=0
      self.af.cur_points=[]
      self.af.cur_index=0
      self.af.step_size=self.slider_frame.getStepValue()
      self.af.cur_zoom=self.zoom
      self.af.next_pos=(self.slider_frame.getValue(), 0)        # Startposition holen und setzen
      self.af.next_posLL=self.trackdata[self.af.track_handle].track[self.af.next_pos[0]][0]
      self.osm.setCenter(self.af.next_posLL, self.handle)
      self.Refresh()
      self.on_track=True


  # ###########################################################
  # Räumt nach Autofollow auf.
  def finishAutofollow(self):
    if not self.autofollow_is_running:
      return
    self.timer_autofollow.Stop()
    self.autofollow_is_running=False
    if self.handle!=self.af.OSMscr_handles[0]:
      # wenn gerade das zweit-Handle aktiv war: dessen Mittelpunkt holen und für das
      # Haupt-Handle einstellen - sonst gäbe es Map-Hüpfer.
      centerLL=self.osm.getLatLonForPixel((self.dc_sizeXY[0]/2, self.dc_sizeXY[1]/2), self.af.OSMscr_handles[1])
      self.osm.endMoveMap(self.af.OSMscr_handles[0])
      self.osm.setCenter(centerLL, self.af.OSMscr_handles[0])
      self.osm.drawMap(self.af.OSMscr_handles[0])
      self.handle=self.af.OSMscr_handles[0]
      self.Refresh()
    self.osm.dropMap(self.af.OSMscr_handles[1])
    self.osm.endMoveMap(self.handle)
    self.af.cur_index=0
    self.af.cur_points=[]
    self.track_point_data=None
    self.on_track=False
    try:  # falls der Slider schon geschlossen ist -> Fehler ignorieren
      self.slider_frame.setAutofollow(False)
    except:
      pass


  # ###########################################################
  # Pro Timer-Schritt wird ein Schritt im Autofollow
  # ausgeführt.
  # Dabei wird immer zwischen zwei Maps hin- und hergeschaltet.
  # Auf der aktiven Map wird der Track per Shift bis zum
  # nächsten NewMapPoint abgefahren. Parallel dazu wird (per
  # Thread in OSMscr3) die zweite Map berechnet, deren Zentrum
  # genau auf dem NewMapPoint liegt.
  def onTimerAutofollow(self, evt):
    if self.af.cur_zoom!=self.zoom:                               # wenn der Zoom-level geändert wurde
      self.finishAutofollow()                                     # Autofollow stoppen und neu starten
      self.prepareAutofollow()
      self.autofollow_is_running=True
      self.timer_autofollow.Start(self.timer_autofollow_interval)
      self.slider_frame.setAutofollow(True)
      return
    if self.esc or not self.on_track:                             # wenn die ESC-Taste gedrückt oder die Map verschoben wurde
      self.finishAutofollow()                                     # wird Autofollow beendet
      return
    if self.af.cur_index==0:                                      # wenn ein neuer Track-Abschnitt beginnt
      rc=self.osm.getTrackPartAsPixelDelta(self.af.track_handle, self.af.next_pos[0], self.af.next_pos[1])
      self.af.next_pos, self.af.next_posLL, self.af.cur_points=rc # Folgeposition und Strecke dahin bestimmen und merken
      if self.af.next_pos is None:                                # wenn keine Folgeposition existiert
        self.finishAutofollow()                                   # ist der Track am letzten Punkt angekommen
        return
      self.osm.setCenter(self.af.next_posLL, self.handle^1)       # zweite Map vorbereiten
      self.osm.drawMap(self.handle^1)
    if self.af.cur_index<len(self.af.cur_points):                 # wenn der Track-Abschnitt noch nicht vollständig abgefahren wurde
      dxy, track_index=self.af.cur_points[self.af.cur_index]      # nächsten Zwischenschritt holen
      self.af.cur_index+=self.af.step_size                        # Index für den nächsten Zwischenschritt einstellen
      self.track_point_data=self.osm.getTrackPointData(self.af.track_handle, track_index) # Daten für diesen Zwischenschritt lesen
      self.osm.doMoveMap(dxy, self.handle)                        # Map per Shift auf diesen Zwischenschritt ausrichten
      self.slider_frame.setValue(track_index)                     # und den Slider nachführen
      self.Refresh()
      wx.Yield()
    else:                                                         # wenn der Track-Abschnitt abgefahren wurde
      self.osm.endMoveMap(self.handle)                            # Shift beenden
      self.af.cur_index=0
      self.af.cur_map_handle^=1                                   # und auf die andere Map...
      self.handle=self.af.OSMscr_handles[self.af.cur_map_handle]  # ...umschalten


  # ###########################################################
  # Stellt das Kontext-Menue dar.
  def onContextMenu(self, evt):
    self.createMenu()
    self.PopupMenu(self.menue)
    self.Refresh()


  # ###########################################################
  # Legt das Kontext-Menue an.
  def createMenu(self):
    self.menue=wx.Menu()
    self.menue.Append(400, 'position to clipboard')
    self.Bind(wx.EVT_MENU, self.copyPosToClipboard, id=400)
    self.menue.AppendSeparator()
    if wardriving_db_found:
      self.menue.Append(500, 'wardriving mode', "", True)
      self.Bind(wx.EVT_MENU, self.wardrivingMode, id=500)
      self.menue.Check(500, self.wardriving_mode)
    self.menue.Append(520, 'reload visible tiles')
    self.Bind(wx.EVT_MENU, self.reloadTiles, id=520)
    if DEBUG:
      self.submenue_debug=wx.Menu()
      self.menue.AppendSubMenu(self.submenue_debug, 'debug')
      self.submenue_debug.Append(550, "show grid for tracks (on tile rebuild)", "", True)
      self.submenue_debug.Check(550, self.menu_show_grid)
      self.Bind(wx.EVT_MENU, self.menuShowGrid, id=550)
      self.submenue_debug.Append(552, "show track points (on load)", "", True)
      self.Bind(wx.EVT_MENU, self.menuShowTrackPoints, id=552)
      self.submenue_debug.Check(552, self.menu_show_track_points)
      self.submenue_debug.Append(554, "show newMap points (on wheel)", "", True)
      self.Bind(wx.EVT_MENU, self.menuShowNewMapPoints, id=554)
      self.submenue_debug.Check(554, self.menu_show_new_map_points)
    
    if len(self.trackdata)>0:       # wenn mindestens ein Track geladen ist
      self.menue.AppendSeparator()
      self.submenue_tracks1=wx.Menu()
      self.menue.AppendSubMenu(self.submenue_tracks1, 'hilite track')
      self.submenue_tracks1.Append(2500, "no track")
      self.Bind(wx.EVT_MENU, self.trackMenu1, id=2500)
      self.submenue_tracks2=wx.Menu()
      self.menue.AppendSubMenu(self.submenue_tracks2, 'remove track')
      self.submenue_tracks2.Append(2000, "all tracks")
      self.Bind(wx.EVT_MENU, self.trackMenu2, id=2000)
      for th, data in self.trackdata.items():
        t1=datetime.fromtimestamp(data.start_ts).strftime("%Y.%m.%d %H:%M:%S")
        t2=datetime.fromtimestamp(data.end_ts).strftime("%H:%M:%S")
        fn=os.path.basename(data.filename)
        self.submenue_tracks1.Append(2501+th, "%03d: %s - %s (%5.1fkm)"%(th, t1, t2, data.dist), fn)
        self.Bind(wx.EVT_MENU, self.trackMenu1, id=2501+th)
        self.submenue_tracks2.Append(2001+th, "%03d: %s - %s (%5.1fkm)"%(th, t1, t2, data.dist), fn)
        self.Bind(wx.EVT_MENU, self.trackMenu2, id=2001+th)
      self.menue.Append(530, 'select track from screen')
      self.Bind(wx.EVT_MENU, self.selectTrackFromScreen, id=530)
      self.menue.Append(540, 'show slider\tS')
      self.Bind(wx.EVT_MENU, self.showSliderFrame, id=540)
      self.submenue_tracks1.Bind(wx.EVT_MENU_HIGHLIGHT, self.menu_highlight)
      self.submenue_tracks2.Bind(wx.EVT_MENU_HIGHLIGHT, self.menu_highlight)


  # ###########################################################
  # Debugging-Funktionen
  def menuShowGrid(self, evt):
    self.menu_show_grid=not self.menu_show_grid
    self.osm.setShowGrid(self.menu_show_grid)
    self.Refresh()
  def menuShowTrackPoints(self, evt):
    self.menu_show_track_points=not self.menu_show_track_points
  def menuShowNewMapPoints(self, evt):
    self.menu_show_new_map_points=not self.menu_show_new_map_points
    self.Refresh()


  # ###########################################################
  # Setzt den Dateinamen des aktuell markierten Menü-Eintrages
  # in der Statusbar.
  def menu_highlight(self, evt):
    if evt.GetId()>0:
      fn=self.menue.GetHelpString(evt.GetId())
      self.statusbar.SetStatusText("%s"%(fn,), 4)


  # ###########################################################
  # Hebt den Track gemäß "evt" hervor.
  def trackMenu1(self, evt):
    td_idx=evt.GetId()-2501
    self.finishAutofollow()               # Autofollow stoppen
    if td_idx<0:                          # "no track" wurde gewählt
      if self.hilite_track_handle>=0:     # wenn derzeit ein Track hervorgehoben ist
        self.osm.setTrackPen(self.hilite_track_handle, (199, 21, 133), width=2) # Hervorhebung aus
        self.hilite_track_handle=-1
    else:                                 # wenn ein Track hervorgehoben werden soll
      if self.hilite_track_handle>=0:     # wenn derzeit ein Track hervorgehoben ist
        self.osm.setTrackPen(self.hilite_track_handle, (199, 21, 133), width=2) # Hervorhebung aus
      self.hilite_track_handle=td_idx
      self.osm.setTrackPen(td_idx, (255, 0, 0), width=4)  # # Hervorhebung ein
    self.osm.forceRefresh()
    self.Refresh()


  # ###########################################################
  # Löscht den Track gemäß "evt".
  def trackMenu2(self, evt):
    td_idx=evt.GetId()-2001                             # Track-Handle entspricht EventID
    self.finishAutofollow()                             # Autofollow stoppen
    if td_idx<0:                                        # alle Tracks löschen gewählt
      for track_handle, data in self.trackdata.items(): # über alle Tracks
        self.osm.unloadTrack(track_handle)              # Track aus der Map löschen
        self.osm.unloadPoints(data.pause_points_handle) # seine zugehörigen Punkte auch
      self.trackdata=dict()                             # und hier alle zusammen löschen
      self.closeSliderFrame()                           # den SliderFrame braucht es jetzt nicht mehr
    else:                                               # ein einzelner Track soll gelöscht werden
      self.osm.unloadTrack(td_idx)                      # Track aus der Map löschen
      self.osm.unloadPoints(self.trackdata[td_idx].pause_points_handle) # seine Punkte ebenfalls
      del self.trackdata[td_idx]                        # und noch hier löschen
      if len(self.trackdata)==0:                        # wenn der letzte Track gelöscht wurde...
        self.closeSliderFrame()                         # muss ein ggf. offener Slider-Frame weg
      else:                                             # ansonsten ist der Slider ggf. auf den nächsten Track einzustellen
        if self.slider_track_handle==td_idx:            # wenn der Track im Slider gelöscht wurde
          tdh=min(self.trackdata)                       # nächsten Track auswählen
          self.loadTrackIntoSlider(tdh)                 # und in den Slider laden
          self.statusbar.SetStatusText("", 3)
          self.statusbar.SetStatusText("", 4)
    self.osm.forceRefresh()
    self.Refresh()


  # ###########################################################
  # Kopiert self.rightDownPosXY ins Clipboard.
  def copyPosToClipboard(self, evt):
    self.copyToClipboard("%14.12f, %14.12f"%self.osm.getLatLonForPixel(self.rightDownPosXY))


  # ###########################################################
  # Kopiert "text" ins Clipboard.
  def copyToClipboard(self, text):
    if wx.TheClipboard.Open():
      do=wx.TextDataObject()
      do.SetText(text)
      wx.TheClipboard.SetData(do)
      wx.TheClipboard.Close()
    else:
      wx.MessageBox(u"Unable to open clipboard!", "Error", wx.OK|wx.ICON_ERROR)


  # ###########################################################
  # Wird aus dem Menü aufgerufen und versucht, die
  # Klick-Position auf einen Track zu mappen. Wenn genau ein
  # passender Track erkannt wurde, wird dieser in den Slider
  # geladen.
  def selectTrackFromScreen(self, evt):
    p=self.rightDownPosXY
    track_handles=self.osm.findTrackForPoint(p, self.handle)
    l=len(track_handles)
    if l==0:
      self.infoText=(p, " no track found ")
      self.Refresh()
    elif l>1:
      self.infoText=(p, " %d tracks found "%(l,))
      self.Refresh()
    else: # ==1
      self.infoText=(p, " found track %03d "%(track_handles[0],))
      self.Refresh()
      pt=self.osm.getLatLonForPixel(p, self.handle)
      idx=self.osm.findTrackPointIndexForPoint(pt, track_handles[0])
      self.showSliderFrame()
      self.loadTrackIntoSlider(track_handles[0], idx)


  # ###########################################################
  # Schaltet die Anzeige der Wardriving-AP-Koordinaten um bzw.
  # an oder aus.
  def wardrivingMode(self, evt):
    if not wardriving_db_found:
      print "wardriving-DB not found!"
      return
    self.finishAutofollow()   # Autofollow stoppen
    self.wardriving_mode=not self.wardriving_mode
    if self.wardriving_mode:
      print "wardrivingMode"
      aps=self.db.getAllAPs()
      self.point_list=list()
      for ap in aps:
        #      0    1     2     3       4       5       6             7           8
        # ap=(lat, lon, bssid, name, channel, crypto, calc_points, first_seen, last_seen)
        if ap[3] not in ["Vodafone Hotspot", "Vodafone Homespot", "Telekom_FON"]:
          data=(ap[2], ap[3], int(ap[4]), ap[5], int(ap[6]), ap[7], ap[8], float(ap[0]), float(ap[1]))
          if ap[5]=="None":
            self.point_list.append(((float(ap[0]), float(ap[1])), 5, (0, 128, 0), (0, data)))
          else:
            self.point_list.append(((float(ap[0]), float(ap[1])), 3, (255, 0, 0), (0, data)))
      self.wardriving_point_handle=self.osm.loadPoints(self.point_list)
    else:   # der Wardriving-Mode war an und wurde gerade abgeschaltet
      self.osm.unloadPoints(self.wardriving_point_handle)
      self.wardriving_point_handle=None
      self.removeAPSeenAtList()   # ggf. eine alte Sichtungs-Liste entfernen
    self.osm.forceRefresh()
    self.Refresh()


  # ######################################################################
  # Löscht eine alte Sichtungs-Liste.
  def removeAPSeenAtList(self):
    if self.handle_for_ap_list is not None:   # wenn überhaupt eine alte Sichtungs-Liste existiert
      self.osm.unloadPoints(self.handle_for_ap_list[0])  # weg damit
      self.osm.unloadPoints(self.handle_for_ap_list[1])
      self.ap_seen_at_list=None
      self.handle_for_ap_list=None


  # ######################################################################
  # Lädt die aktuell angezeigten Tiles (inkl. Rand-Tiles) neu vom
  # Tile-Server.
  def reloadTiles(self, evt):
    self.finishAutofollow()
    self.osm.forceRefresh()
    self.osm.drawMap(force_reload=True)
    self.Refresh()


  # ######################################################################
  # Stellt den Slider für den Track "track_handle" ein.
  def loadTrackIntoSlider(self, track_handle, index=0):
    if track_handle in self.trackdata:                  # wenn das Track-Handle existiert
      if self.slider_frame is not None:                 # wenn der SliderFrame offen ist
        self.slider_track_handle=track_handle           # Handle zum Slider merken
        data=self.trackdata[track_handle]
        self.slider_frame.setTitle(data.filename)       # und Slider auf Track einstellen
        self.slider_frame.setMaxVal(len(data.track)-1)
        self.slider_frame.setValue(index)


  # ######################################################################
  # Menü-Funktion "show slider" wurde gewählt - SliderFrame öffnen.
  def showSliderFrame(self, evt=None):
    if len(self.trackdata)==0:        # wenn keine Tracks geladen sind
      return                          # gibts auch keinen Slider
    if self.slider_frame is not None: # wenn Slider-Frame schon offen ist -> nur nach vorne holen
      self.slider_frame.Iconize(False)
      return
    self.slider_queue=Queue()
    #pos=self.getSliderFramePos()
    pos=wx.DefaultPosition
    size=(self.dc_sizeXY[0], -1)
    self.slider_frame=SliderFrame(self, "Track-Slider", pos, size, self.slider_queue)


  # ######################################################################
  # SliderFrame schließen und zugehörige Variablen zurücksetzen.
  def closeSliderFrame(self, is_already_closed=False):
    if self.slider_frame is not None and not is_already_closed:
      self.slider_frame.Close()
      wx.Yield()
    self.slider_frame=None
    self.slider_queue=None
    self.slider_track_handle=None







# ###########################################################
# Nur ein Fenster mit einem Slider. Die Rückmeldung von
# Änderungen der Slider-Position erfolgt via Queue.
#
# Vorüberlegungen bzgl. Pixel-Schrittweite und Zweit-Map-Nutzung:
# - Der Track sollte in Tilesize lange Abschnitte zerlegt werden.
# - Diese Abschnitte sollten immer per doMoveMap() abgefahren werden.
# - Bei kleinen Zoom-Leveln (z.B. <=14) müssen ggf. Trackpunkte übersprungen werden, um pro doMoveMap() eine
#   gleichbleibende Anzahl von Pixeln zu erhalten.
# - Der Zoom-Level muss sich auch bei laufender Nachfahrt ändern lassen.
# - Pause-Punkte sollten "übersprungen" werden.
# - Wird das automatische Nachfahren gestoppt, sollte die Position des Sliders nachgeführt werden, wenn
#   der Karten-Ausschnitt verändert wird (und der Track im Ausschnitt liegt).
# - (Die Anzahl der möglichen Slider-Positionen sollte auch bei Zoom-Level-Änderung konstant bleiben.)
#
# - Bei mehreren geladenen Tracks muss sich der nachzufahrende Track auswählen lassen.
# - Scroll-Schritt-Frequenz und Schrittweite müssen einstellbar sein.
#
# -> Der Track könnte pro Zoom-Level so umgerechnet werden, dass er in Pixel-Schrittweite vorliegt.
#     Parallel dazu könnte es eine Liste geben, die die Tile-Grenzen kennzeichnet.
#     Allerdings würde das wohl ziemlich lange Listen ergeben.
# -> Vielleicht könnten auch die Track-Daten der Tiles herangezogen werden.
#     Berechnung der Zwischenschritte dann ggf. on-the-fly pro Tile.
#     Bei Zoom-Wechsel wird dictTN pro Tile ohnehin als erstes berechnet....
#     Andererseits ergäbe sich da die Frage nach dem Folge-Tile und die aus dem Tile herausragenden
#     Trackpunkte dürften ebenfalls unerfreulich werden.
#
# Template von: https://www.tutorialspoint.com/wxpython/wx_slider_class.htm
class SliderFrame(wx.Frame):
  def __init__(self, parent, title, pos, size, queue):
    super(SliderFrame, self).__init__(parent, title=title, pos=pos) #, size=size)
    self.max_val=1
    self.queue=queue
    miny=self.InitUI()
    self.SetClientSize((size[0], miny)) # die Frame-Höhe dynamisch einstellen


  # ###########################################################
  # Den Fenster-Inhalt anlegen und Event-Routinen aufsetzten.
  def InitUI(self):
    panel=wx.Panel(self)
    self.sld=wx.Slider(panel, value=0, minValue=0, maxValue=self.max_val, style=wx.SL_HORIZONTAL|wx.SL_LABELS)
    self.sld_last_value=-1  # zur Ausblendung von Duplikaten

    txt_freq=wx.StaticText(panel, label="freq[ms]")
    self.spn_freq=wx.SpinCtrl(panel, size=(50, -1), min=1, max=100)
    self.spn_freq.SetValue(10)
    vbox_freq=wx.BoxSizer(wx.VERTICAL)
    vbox_freq.Add(txt_freq,      0, flag=wx.ALIGN_CENTER_HORIZONTAL|wx.ALL, border=0)
    vbox_freq.Add(self.spn_freq, 0, flag=wx.ALIGN_CENTER_HORIZONTAL|wx.ALL, border=0)

    txt_step=wx.StaticText(panel, label="step[px]")
    self.spn_step=wx.SpinCtrl(panel, size=(50, -1), min=1, max=10)
    self.spn_step.SetValue(1)
    vbox_step=wx.BoxSizer(wx.VERTICAL)
    vbox_step.Add(txt_step,      0, flag=wx.ALIGN_CENTER_HORIZONTAL|wx.ALL, border=0)
    vbox_step.Add(self.spn_step, 0, flag=wx.ALIGN_CENTER_HORIZONTAL|wx.ALL, border=0)

    self.btn=wx.ToggleButton(panel, label=" Auto\nfollow", size=(50, -1))

    hbox=wx.BoxSizer(wx.HORIZONTAL)
    hbox.Add(self.sld,  1, flag=wx.EXPAND|wx.ALL, border=1)
    hbox.Add(vbox_freq, 0, flag=wx.ALIGN_CENTER_VERTICAL|wx.ALL, border=1)
    hbox.Add(vbox_step, 0, flag=wx.ALIGN_CENTER_VERTICAL|wx.ALL, border=1)
    hbox.Add(self.btn,  0, flag=wx.ALIGN_CENTER_VERTICAL|wx.ALL, border=1)

    self.Bind(          wx.EVT_CLOSE,         self.onClose)
    self.sld.Bind(      wx.EVT_SCROLL,        self.onScroll)
    self.spn_freq.Bind( wx.EVT_SPINCTRL,      self.onSpinFreq)
    self.spn_step.Bind( wx.EVT_SPINCTRL,      self.onSpinStep)
    self.btn.Bind(      wx.EVT_TOGGLEBUTTON,  self.onToggle)
    panel.SetSizer(hbox)
    self.Show(True)
    return(self.spn_step.GetSize()[1]+txt_step.GetSize()[1]+4)


  # ###########################################################
  # Wird beim Schließen des Frames aufgerufen, schreibt "CLOSE"
  # auf die Queue und schließt den Frame.
  def onClose(self, evt):
    self.queue.put(("CLOSE", 0))
    self.Destroy()


  # ###########################################################
  # Wird bei Änderung der Slider-Position aufgerufen und
  # schreibt die aktuelle Slider-Position auf die Queue.
  def onScroll(self, evt):
    value=self.sld.GetValue()
    if value!=self.sld_last_value:  # Duplikate sind lästig
      self.queue.put(("SCROLL", value))
      self.sld_last_value=value


  # ###########################################################
  # Wird bei Änderung des Frequenz-Wählers aufgerufen und
  # schreibt dessen Wert auf die Queue.
  def onSpinFreq(self, evt):
    self.queue.put(("FREQ", self.spn_freq.GetValue()))


  # ###########################################################
  # Wird bei Änderung des Schrittweiten-Wählers aufgerufen und
  # schreibt dessen Wert auf die Queue.
  def onSpinStep(self, evt):
    self.queue.put(("STEP", self.spn_step.GetValue()))


  # ###########################################################
  # Wird bei Änderung des Zustands bei Autofollow aufgerufen
  # und schreibt dessen Wert auf die Queue.
  def onToggle(self, evt):
    self.queue.put(("AUTO", int(self.btn.GetValue()==True)))


  # ###########################################################
  # Setzt den Maximal-Wert des Sliders auf "max_val".
  def setMaxVal(self, max_val):
    self.sld.SetRange(0, max_val)
    self.max_val=max_val


  # ###########################################################
  # Setzt die Position des Sliders auf einen Wert zwischen
  # 0 und "self.max_val".
  def setValue(self, val):
    if 0<=val<=self.max_val:
      self.sld.SetValue(val)


  # ###########################################################
  # Liefert die aktuelle Position des Sliders.
  def getValue(self):
    return(self.sld.GetValue())


  # ###########################################################
  # Setzt den Fenster-Titel auf "title".
  def setTitle(self, title):
    self.SetTitle(title)


  # ###########################################################
  # Liefert den aktuellen Wert der Schrittgröße.
  def getStepValue(self):
    return(self.spn_step.GetValue())


  # ###########################################################
  # Schaltet den Autofollow-Button ein (onoff=True) oder aus.
  def setAutofollow(self, onoff=False):
    self.btn.SetValue(onoff)



# ###########################################################
# Der Fenster-Rahmen für das Hauptfenster.
class MapFrame(wx.Frame):
  def __init__(self, parent, centerLL, zoom, border, title, pos=wx.DefaultPosition, size=wx.DefaultSize):
    style=wx.DEFAULT_FRAME_STYLE
    wx.Frame.__init__(self, None, wx.ID_ANY, title+" "+VERSION, pos=pos, size=size, style=style)
    win=MapWindow(self, centerLL, zoom, border, title, size)





# ######################################################################
# main()
if __name__=="__main__":
  pt=(54.805060, 9.524878)
  
  app=wx.App(False)
  frame=MapFrame(None, pt, 16, (2, 2), "wxOSM", size=(1280, 768+25))  # 5x3 Tiles + 25 für Statusbar
  frame.Show()
  app.MainLoop()
